#include <errno.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

#include <zephyr/kernel.h>
#include <zephyr/arch/cpu.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/logging/log.h>
#include <zephyr/sys/util.h>

#include <zephyr/device.h>
#include <zephyr/init.h>
#include <zephyr/drivers/uart.h>

#include <zephyr/net/buf.h>


/* size of stack area used by each thread */
#define STACKSIZE 1024

#define LOG_MODULE_NAME pivot_bmd
LOG_MODULE_REGISTER(LOG_MODULE_NAME);



typedef enum {
	CMD_GET_VERSION = 0x05,
	CMD_CHECK_MOBILE = 0x0C,
    CMD_BLE_REQUEST_RESET = 0x10,       // BLE device request PIC reset        
} _tBMD_CMD_CODE;

typedef enum {
    MOBILE_CHECK_NOTHING = 0,           // Catch all unknown cases                
    MOBILE_CHECK_FOUND = 1,             // Found a mobile access credential
    MOBILE_CHECK_NOT_FOUND = 2,         // Could not find a mobile access credential
    MOBILE_CHECK_PENDING = 3,           // Still looking for mobile access credential
    MOBILE_CHECK_COMPLETE = 5,          // Status reported to phone
                
    MOBILE_CHECK_SCANNING = 6,
    MOBILE_CHECK_CONNECTING = 4,        // UUID match and in range
    MOBILE_CHECK_CONNECTED = 7,
    MOBILE_CHECK_DISCOVERIED = 8,       // Service found
    MOBILE_CHECK_DISCONNECTED = 9                
} _tMOBILE_CHECK_STATE;

#define RESP_CODE_PASS	0x00
#define RESP_CODE_FAIL	0x01

static const struct device *const pic_uart_dev = DEVICE_DT_GET(DT_NODELABEL(uart0));

static K_THREAD_STACK_DEFINE(tx_thread_stack, STACKSIZE);
static struct k_thread tx_thread_data;

static K_FIFO_DEFINE(uart_tx_queue);

#define MAX_PAYLOAD_SIZE 32-3           // 0x7E, len, CMD
#define PKT_HEADER 0x7E

/* Receiver states. */
#define ST_IDLE 0	/* Waiting for packet type. */
#define ST_HDR 1	/* Receiving packet header. */
#define ST_PAYLOAD 2	/* Receiving packet payload. */
#define ST_DISCARD 3	/* Dropping packet. */

#define DISCARD_LEN 33


struct uart_pkt_hdr {
	uint8_t length;			// 2 + payload size
	uint8_t cmd;
} __packed;

static uint8_t CalcXOR(uint8_t *XorData, uint16_t Size) {
    uint8_t xorVal;
    uint8_t i;

    xorVal = XorData[0];
    for (i = 1; i < Size; i++) {
        xorVal ^= XorData[i];
    }

    return xorVal;
}


static int uart_read(const struct device *uart, uint8_t *buf, size_t len)
{
    int rx = uart_fifo_read(uart, buf, len);

	LOG_DBG("read %d req %d", rx, len);

	return rx;
}

static void ReadUartPacket(void)
{
	static int remaining;
	static uint8_t state;
	static uint8_t data;
    static uint8_t cmd;
	static uint8_t packet_buf[32];
	static uint8_t data_buf[32];
	int read;

	do {
		switch (state) {
		case ST_IDLE:
			/* Get packet type */
			read = uart_read(pic_uart_dev, &data, sizeof(data));
			/* since we read in loop until no data is in the fifo,
			 * it is possible that read = 0.
			 */
			if (read) {
				if (data == PKT_HEADER) {
					/* Get expected header size and switch
					 * to receiving header.
					 */
					remaining = sizeof(struct uart_pkt_hdr);
					state = ST_HDR;
				} else {
					LOG_WRN("Unknown header %d", data);
				}
			}
			break;
		case ST_HDR:
			read = uart_read(pic_uart_dev, &packet_buf[sizeof(struct uart_pkt_hdr) - remaining], remaining);
			remaining -= read;
			if (remaining == 0) {
				/* Header received. Allocate buffer and get
				 * payload length. If allocation fails leave
				 * interrupt. On failed allocation state machine
				 * is reset.
				 */
				remaining = packet_buf[0] - sizeof(struct uart_pkt_hdr);

				if (remaining > MAX_PAYLOAD_SIZE) {
					LOG_ERR("Not enough space in packet buffer");
					state = ST_DISCARD;
				} else {
					state = ST_PAYLOAD;
				}

			}
			break;

		case ST_PAYLOAD:
			read = uart_read(pic_uart_dev, &packet_buf[packet_buf[0] - remaining], remaining);
			remaining -= read;
			if (remaining == 0) {
				/* Packet received */
				LOG_DBG("putting RX packet in queue.");

				if(packet_buf[packet_buf[0]] == CalcXOR(packet_buf, packet_buf[0])) {
                    cmd = packet_buf[1];
                    switch(cmd) {
                        case CMD_CHECK_MOBILE:
                            // Send response to PIC
							data_buf[0] = 6;
							data_buf[1] = PKT_HEADER;
							data_buf[2] = 4;
							data_buf[3] = CMD_CHECK_MOBILE;
							data_buf[4] = RESP_CODE_PASS;	
							data_buf[5] = MOBILE_CHECK_PENDING;
							data_buf[6] = CalcXOR(&data_buf[2], 4);	

							char *mem_ptr = k_malloc(7);
							__ASSERT_NO_MSG(mem_ptr != 0);
							memcpy(mem_ptr, data_buf, 7);
							k_fifo_put(&uart_tx_queue, mem_ptr);

                            state = ST_IDLE;
                            break;
                        default:
                            break;        
                    }
                } else {
                    state = ST_DISCARD;
            	}				
			}
			break;
		case ST_DISCARD:
		{
			uint8_t discard[32];
			size_t to_read = MIN(remaining, sizeof(discard));

			read = uart_read(pic_uart_dev, discard, to_read);
			remaining -= read;
			if (remaining == 0) {
				state = ST_IDLE;
			}

			break;

		}
		default:
			read = 0;
			__ASSERT_NO_MSG(0);
			break;

		}
	} while (read);
}

void uart_tx_thread(void)
{
	uint8_t *data_buf;

	while(1) {
		data_buf = k_fifo_get(&uart_tx_queue, K_FOREVER);
		uart_fifo_fill(pic_uart_dev, &data_buf[1], data_buf[0]);
		k_free(data_buf);

		k_yield();
	}
}

static void uart_rx_isr(const struct device *unused, void *user_data)
{
	ARG_UNUSED(unused);
	ARG_UNUSED(user_data);

	if (!uart_irq_rx_ready(pic_uart_dev)) {
		LOG_DBG("spurious interrupt");
	}

	if (uart_irq_rx_ready(pic_uart_dev)) {
		ReadUartPacket();
	}
}

static int pivot_bmd_init(const struct device *unused)
{
	LOG_DBG("");

	if (!device_is_ready(pic_uart_dev)) {
		LOG_ERR("HCI UART %s is not ready", pic_uart_dev->name);
		return -EINVAL;
	}

	uart_irq_rx_disable(pic_uart_dev);
	uart_irq_tx_disable(pic_uart_dev);

	uart_irq_callback_set(pic_uart_dev, uart_rx_isr);
	uart_irq_rx_enable(pic_uart_dev);

	return 0;
}


SYS_INIT(pivot_bmd_init, APPLICATION, CONFIG_KERNEL_INIT_PRIORITY_DEVICE);

void main(void)
{
	/* incoming events and data from the PIC */
	static K_FIFO_DEFINE(uart_rx_queue);
	int err;

	LOG_DBG("Start");
	__ASSERT(pic_uart_dev, "UART device is NULL");

	/* Spawn the TX thread and start feeding commands and data to the
	 * PIC
	 */
	k_thread_create(&tx_thread_data, tx_thread_stack,
			K_THREAD_STACK_SIZEOF(tx_thread_stack), uart_tx_thread,
			NULL, NULL, NULL, K_PRIO_COOP(7), 0, K_NO_WAIT);
	k_thread_name_set(&tx_thread_data, "Pivot-BMD uart TX");

	while (1) {

	}
}
